/* cmod - set up the user environment with module files.
   Copyright (C) 1998 by the authors of cmod (see AUTHORS).

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdio.h>

#include "cmod_string.h"
#include "cmod_token.h"
#include "cmod_preproc.h"
#include "archaic.h"
#include "cmod_main.h"
#include "env.h"
#include "cmod_process.h"


enum dollar_state {
  dollar_none,
  dollar_found,			/* found "$" */
  dollar_var,			/* found "${" */
  dollar_script			/* found "$(" */
};

enum inner_state {
  inner_none,
  inner_backslash,
  inner_dollar,
  inner_var
};


static void
prep_error(const char *errmsg,
	   const struct cmod_str *filename,
	   int *lineno)
{
  fprintf(stderr, "%s:%d:%s\n",
	  str_asciz(filename), 1 + *lineno, errmsg);
}


enum cmod_status
cmod_preprocess(FILE *file,                      /* in */
		const struct cmod_str *filename, /* in */
		struct token_state *buffer,      /* in/out */
		int *lineno)                     /* in/out */
{
  int c;
  struct cmod_str shellbuf;
  struct cmod_str varbuf;
  enum dollar_state dollar = dollar_none;
  enum inner_state inner = inner_none;
  struct env *env_var;

  if (buffer->eof_seen)
    return cmod_eof;

  /* FIXME: this could possible be speeded up a trifle by not
     initializing shellbuf and varbuf until they are actually
     needed. */
  str_ctor(&shellbuf);
  str_ctor(&varbuf);

  while ((c = getc(file)) != EOF)
    {
      if (c == '\n')
	++*lineno;

      switch (dollar)
	{
	case dollar_none:
	  if (c == '$')
	    dollar = dollar_found;
	  else
	    {
	      str_append_char(&buffer->buffer, c);
	      if (c == '\n')
		{
		  str_dtor(&shellbuf);
		  str_dtor(&varbuf);
		  return cmod_ok;
		}
	    }
	  break;
	case dollar_found:
	  switch (c)
	    {
	    case '{':
	      dollar = dollar_var;
	      str_clear(&varbuf);
	      break;
	    case '(':
	      dollar = dollar_script;
	      str_clear(&shellbuf);
	      break;
	    case '$':
	      dollar = dollar_none;
	      str_append_char(&buffer->buffer, '$');
	      break;
	    default:
	      prep_error("bad dollar substitution", filename, lineno);
	      goto cleanup;
	    }
	  break;
	case dollar_var:
	  switch (c)
	    {
	    case '}':
	      env_var = envarray_locate_create(global_env, &varbuf, 1, 0);
	      if (env_var == NULL)
		{
		  fprintf(stderr, "%s:%d:unbound variable %s\n",
			  str_asciz(filename), 1 + *lineno,
			  str_asciz(&varbuf));
		  goto cleanup;
		}

	      /* FIXME: what do we do with explicitly unset variables?
                 Think hard about unloading modules... */
	      env_value(env_var, &varbuf);
	      str_append_str(&buffer->buffer, &varbuf);
	      dollar = dollar_none;
	      if (str_strchr(&buffer->buffer, '\n') != -1)
		{
		  str_dtor(&shellbuf);
		  str_dtor(&varbuf);
		  return cmod_ok;
		}
	      break;
	    case '\n':
	      prep_error("unterminated variable reference", filename, lineno);
	      goto cleanup;
	    default:
	      str_append_char(&varbuf, c);
	      break;
	    }
	  break;
	case dollar_script:
	  switch (inner)
	    {
	    case inner_none:
	      switch (c)
		{
		case '\\':
		  inner = inner_backslash;
		  break;
		case '$':
		  inner = inner_dollar;
		  break;
		case ')':
		  str_clear(&varbuf);
		  if (cmod_popen(&shellbuf, &varbuf) != 0)
		    goto cleanup;

		  /* Remove a trailing newline, if present. */
		  if (str_tail(&varbuf) == '\n')
		    str_remove_last(&varbuf, 1);

		  str_append_str(&buffer->buffer, &varbuf);
		  dollar = dollar_none;
		  if (str_strchr(&buffer->buffer, '\n') != -1)
		    {
		      str_dtor(&shellbuf);
		      str_dtor(&varbuf);
		      return cmod_ok;
		    }
		  break;
		default:
		  str_append_char(&shellbuf, c);
		  break;
		}
	      break;
	    case inner_backslash:
	      switch (c)
		{
		case '$':
		case '(':
		case '\\':
		case ')':
		  str_append_char(&shellbuf, '$');
		  inner = inner_none;
		  break;
		case '\n':
		  inner = inner_none;
		  break;
		default:
		  prep_error("bad backslash escape in shell escape",
			     filename, lineno);
		  goto cleanup;
		}
	      break;
	    case inner_dollar:
	      if (c == '{')
		{
		  str_clear(&varbuf);
		  inner = inner_var;
		}
	      else
		{
		  prep_error("bad dollar usage in shell escape",
			     filename, lineno);
		  goto cleanup;
		}
	      break;
	    case inner_var:
	      switch (c)
		{
		case '}':
		  env_var = envarray_locate_create(global_env, &varbuf, 1, 0);
		  if (env_var == NULL)
		    {
		      fprintf(stderr, "%s:%d:unbound variable %s\n",
			      str_asciz(filename), 1 + *lineno,
			      str_asciz(&varbuf));
		      goto cleanup;
		    }

		  /* FIXME: what do we do with explicitly unset variables?
		     Think hard about unloading modules... */
		  env_value(env_var, &varbuf);
		  str_append_str(&shellbuf, &varbuf);
		  inner = inner_none;
		  break;
		case '\n':
		  prep_error("unterminated variable reference in shell escape",
			     filename, lineno);
		  goto cleanup;
		default:
		  str_append_char(&varbuf, c);
		  break;
		}
	      break;
	    }
	}
    }
  
  buffer->eof_seen = 1;
  switch (dollar)
    {
    case dollar_none:
      break;
    case dollar_found:
      prep_error("unexpected end of file after dollar sign",
		 filename, lineno);
      goto cleanup;
    case dollar_var:
      prep_error("unexpected end of file in variable reference",
		 filename, lineno);
      goto cleanup;
    case dollar_script:
      prep_error("unexpected end of file in shell escape",
		 filename, lineno);
      goto cleanup;
      
    }

  str_dtor(&shellbuf);
  str_dtor(&varbuf);
  
  return str_empty(&buffer->buffer) ? cmod_eof : cmod_ok;

cleanup:
  str_dtor(&shellbuf);
  str_dtor(&varbuf);
  return cmod_abort_module;
}
